///Contains everything needed to play music.
///The player is akin to a virtual cassette player, where the audio files would be cassettes themselves, 
///stored inside the playlist.
pub mod audioplayer {

    use std::fs::File;
    use std::io::BufReader;
    use std::path::{Path, PathBuf};
    use std::time::Duration;
    use rodio::{Decoder, OutputStream, Sink, Source};
    use std::error::Error;
    use std::fmt;
    use crate::playlist::playlist::*;
    use crate::songmeta::songmeta::*;
    use std::io::{Read, Seek};

    //-----------------------------------Errors
    #[derive(Debug, Clone)]
    pub enum PlayerError {
        Something,
    }

    ///Main error container for the player.
    impl fmt::Display for PlayerError {
        fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
            match self {
                PlayerError::Something => write!(f, "Uh-oh"),
            }
        }
    }

    impl Error for PlayerError {}


    //-----------------------------------------Structs
    pub struct AudioPlayer {
        ///A facility for controlling the playing of sound.
        sink: Sink,
        ///A handle to the sound device. Must live as long as the sink.
        stream: OutputStream,
        ///List of songs to pick from.
        pub playlist: Playlist,
        ///If true, the player will loop back to the beginning of the playlist when it's done
        pub repeating: bool,

    }

    //-----------------------------------Enums
    ///Returned by player.get_state()
    #[derive(Debug, Clone)]
    pub enum PlayerState {
        Playing,
        Paused,
        Stopped,
    }

    //-------------------------------------Impls




    impl AudioPlayer {

        ///Returns a new empty player to the default device.
        pub fn new() -> Result<Self, Box<dyn Error>> {
            let (stream, stream_handle) = OutputStream::try_default()?;
            let sink = Sink::try_new(&stream_handle)?;
            let playlist = Playlist::new();

            let me = Self {
                sink: sink,
                stream: stream,
                playlist: playlist,
                repeating: false,
            };
            Ok(me)
        }

        ///Plays all files in the playlist starting from index.
        pub fn play_at(&mut self, index: usize) -> Result<(), Box<dyn Error>> {
            self.playlist.check_index(index.clone())?;
            self.sink.clear();
            for song in &self.playlist.songs[index..] {
                let file = BufReader::new(File::open(&song.path)?);
                let source = Decoder::new(file)?;
                self.sink.append(source);
            }
            self.sink.play();
            return Ok(());
        }

        ///Returns a copy of the current song and its index in the playlist
        ///Or None if the playlist's empty
        pub fn get_current_song(&self) -> Option<(SongMeta, usize)> {
            if self.playlist.songs.len() == 0 ||
            self.sink.len() == 0 {
                return None;
            }
            let index: usize = self.playlist.songs.len() - self.sink.len();
            let songcopy = self.playlist.songs[index].clone();
            return Some((songcopy, index));
        }

        ///Returns the current state of the player
        pub fn get_state(&self) -> PlayerState {
            if self.sink.is_paused() {
                return PlayerState::Paused;
            }
            if self.sink.len() > 0 {
                return PlayerState::Playing;
            }
            return PlayerState::Stopped;
        }

        pub fn pause(&mut self) {
            self.sink.pause();
        }

        pub fn unpause(&mut self) {
            self.sink.play();
        }

        pub fn skip_forward(&mut self) -> Result<(), Box<dyn Error>> {
            unimplemented!();
        }

        pub fn skip_backwards(&mut self) {
            unimplemented!();
        }

        pub fn get_pos(&self) {
            unimplemented!();
        }

        pub fn set_vol(&mut self, vol: f32) {
            unimplemented!();
        }

        pub fn get_vol(&mut self) {
            unimplemented!();
        }

        pub fn goto(&mut self, index: usize) {
            unimplemented!();
        }

        pub fn press_shuffle(&mut self) {
            unimplemented!();
        }

        pub fn press_repeat(&mut self) {
            unimplemented!();
        }

        pub fn seek(&mut self, goto: Duration) {
            unimplemented!();
        }

        pub fn stop(&mut self) {
            unimplemented!();
        }

        pub fn load_file(&mut self) {
            unimplemented!();
        }

        pub fn load_folder(&mut self) {
            unimplemented!();
        }
    }
}
