///Module tied to the SongMeta module.
///Handles storing songmetas into a playlist that is ready to be used by a music player.
///Includes writing and reading from a yaml file for ease-of-use.
///Any functions of a music player that act on its playlist go in here, including sorting and
///shuffling.

pub mod playlist {

    use std::path::{Path,PathBuf};
    use std::error::Error;
    use std::fmt;
    use std::fs;
    use serde::{Deserialize, Serialize};
    use crate::songmeta::songmeta::*;
    use std::io::Write;
    use rand::thread_rng;
    use rand::prelude::SliceRandom;




    //----------------ERROR

    #[derive(Debug)]
    pub enum PlaylistError {
        IndexNotFound,
        FileNotFound,
        FileCantWrite,
    }

    impl fmt::Display for PlaylistError {
        fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
            match self {
                PlaylistError::IndexNotFound => write!(f, "Playlist does not have that index"),
                PlaylistError::FileNotFound => write!(f, "Given playlist file is not readable or does not exist"),
                PlaylistError::FileCantWrite => write!(f, "Can't write playlist file at the given path. Is it valid ?"),
            }
        }
    }

    impl Error for PlaylistError {}


    //----------------STRUCT

    ///Stores a list of SongMetas to be used with a player.
    ///Index starts at zero ! 
    #[derive(Debug, Clone, Serialize, Deserialize)]
    pub struct Playlist {
        pub songs: Vec<SongMeta>,
        pub shuffled: bool,
    }

    impl Playlist {
        ///Returns a new empty playlist.
        pub fn new() -> Self {
            Self {
                songs: vec![],
                shuffled: false,
            }
        }

        ///Returns a playlist from a file.
        ///File must be yaml
        ///If check is true, we remove from the playlist the files that are not present on the
        ///filesystem.
        pub fn from_file(path: &str, check: bool) -> Result<Self, Box<dyn Error>> {

            let p = Path::new(path);

            //Check existence
            if !p.exists() {
                return Err(Box::new(PlaylistError::FileNotFound));
            }

            //Read the yaml
            let content = fs::read_to_string(p)?;
            let mut res: Playlist = serde_yaml::from_str(&content)?;
            
            //Check ?
            if check { res.check_sources(); }
            return Ok(res);
        }

        ///Writes the playlist with its index to a file
        ///Path must be fully qualified
        pub fn to_file(&self, path: &str) -> Result<(), Box<dyn Error>>{
            let p = Path::new(&path);

            //Check parent
            let mut parent = PathBuf::from(path);
            let _ = &parent.pop();
            if !&parent.exists() {
                return Err(Box::new(PlaylistError::FileCantWrite));
            }

            //Serialize and write
            let serialized = serde_yaml::to_string(&self.clone())?;
            let mut file = fs::OpenOptions::new()
                .write(true)
                .append(false)
                .create(true)
                .open(&p)?;
            file.write_all(format!("---\n{}", serialized).as_bytes())?;
            return Ok(());
        }

        ///Checks the existence of every song in the playlist on the filesystem.
        ///Removes the ones that do not exist anymore.
        pub fn check_sources(&mut self) {
            let mut recipient: Vec<SongMeta> = Vec::new(); 
            for song in &self.songs {
                let p = Path::new(&song.path);
                if p.exists() {
                    recipient.push(song.clone());
                }
            }
            self.songs = recipient;
        }


        ///shortcut to myplaylist.songs.push(SongMeta::frompath(&String::from("/a/file/path")))
        pub fn add_song_from_path(&mut self, path: &String) -> Result<(), Box<dyn Error>> {
            let song = SongMeta::frompath(path)?;
            self.songs.push(song);
            return Ok(());
        }

        ///Checks if a number is a valid index in this playlist.
        pub fn check_index(&self, index: usize) -> Result<(), Box<dyn Error>> {
            if self.songs.len() == 0 {
                return Err(Box::new(PlaylistError::IndexNotFound));
            }
            if index > (self.songs.len() - 1) {
                return Err(Box::new(PlaylistError::IndexNotFound));
            }
            return Ok(());
        }

        ///Removes a song at index (if possible)
        pub fn remove(&mut self, index: usize) -> Result<(), Box<dyn Error>> {
            self.check_index(index.clone())?;
            self.songs.remove(index);
            return Ok(());
        }

        ///Shuffles the playlist
        pub fn shuffle(&mut self) {
            //Return if playlist is empty
            if self.songs.len() == 0 {
                return;
            }
            self.songs.shuffle(&mut thread_rng());
            self.shuffled = true;
        }

        ///Sorts the playlist by artist, album, trackorder, title
        pub fn sort(&mut self) {
            self.songs.sort_by(|a, b| {
                a.artist
                    .cmp(&b.artist) 
                    .then_with(|| a.album.cmp(&b.album)) 
                    .then_with(|| a.track.cmp(&b.track)) 
                    .then_with(|| a.title.cmp(&b.title)) 
            });
            self.shuffled = false;
        }
    }
}
